@@ -1,14 +1,12 @@
-/* $EPIC: term.c,v 1.4 2001/12/18 17:04:36 crazyed Exp $ */
+/* $EPIC: term.c,v 1.14 2008/03/18 04:24:01 jnelson Exp $ */
 /*
  * term.c -- termios and (termcap || terminfo) handlers
  *
  * Copyright (c) 1990 Michael Sandroff.
  * Copyright (c) 1991, 1992 Troy Rollo.
  * Copyright (c) 1992-1996 Matthew Green.
- * Copyright  1993, 1997 Jeremy Nelson.
  * Copyright  1998 J. Kean Johnston, used with permission.
- * Copyright  1999 Ben Winslow and others ("EPIC Software Labs").
- * Copyright  1995-2000 Jeremy Nelson and others ("EPIC Software Labs").
+ * Copyright  1995-2002 EPIC Software Labs 
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -36,13 +34,16 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
  * SUCH DAMAGE.
  */
-
+/*
+ * Ben Winslow deserves specific praise for his fine work adding 
+ * terminfo support to EPIC.
+ */
 #define __need_putchar_x__
 #define __need_term_flush__
 #include "irc.h"
 #include "ircaux.h"
 #include "vars.h"
-#include "term.h"
+#include "termx.h"
 #include "window.h"
 #include "screen.h"
 #include "output.h"
@@ -63,33 +64,84 @@
 #endif
 #include <sys/ioctl.h>
 
-	int		need_redraw;
-static	int		tty_des;		/* descriptor for the tty */
-static	struct	termios	oldb, newb;
-	char		my_PC, *BC, *UP;
-	int		BClen, UPlen;
+volatile	int		need_redraw;
+static		int		tty_des;	/* descriptor for the tty */
+static		struct termios	oldb, newb;
 
-/* Systems cant seem to agree where to put these... */
-#ifdef HAVE_TERMINFO
-extern	int		setupterm();
-extern	char		*tigetstr();
-extern	int		tigetnum();
-extern	int		tigetflag();
-#define Tgetstr(x, y) 	tigetstr(x.iname)
-#define Tgetnum(x) 	tigetnum(x.iname);
-#define Tgetflag(x) 	tigetflag(x.iname);
+/*
+ * XXX -- Supporting this correctly is a hack, but it's not my fault.
+ * 
+ * Configure has already determined if <term.h> can be used to get
+ * prototypes for these functions.  It has also determined whether it is 
+ * X/OPEN curses, and therefore requires us to include <termio.h> and 
+ * <curses.h> before including <term.h>.
+ *
+ * For those systems that do not have a usable <term.h>, we will set forth
+ * some empty decl prototypes for the functions, just to establish the 
+ * return value.  All systems have either the terminfo functions or the 
+ * termcap functions, but not all systems have tparm(), so to avoid clashing
+ * with the my_tparm() decl in "compat.h", we only decl tparm() if we are
+ * using the system's (and not our) tparm().
+ */
+#if defined(HAVE_TERM_H) && !defined(DONT_USE_TERM_H)
+# if defined(TERM_H_REQUIRES_CURSES_H)
+#  include <termio.h>
+#  include <curses.h>
+# endif
+# include <term.h>
 #else
-extern	int		tgetent();
-extern	char		*tgetstr();
-extern	int		tgetnum();
-extern	int		tgetflag();
-#define Tgetstr(x, y) 	tgetstr(x.tname, &y)
-#define Tgetnum(x) 	tgetnum(x.tname)
-#define Tgetflag(x) 	tgetflag(x.tname)
+# ifdef HAVE_TERMINFO
+	extern	int	setupterm();
+	extern	char *	tigetstr();
+	extern	int	tigetnum();
+	extern	int	tigetflag();
+# else
+	extern	int	tgetent();
+	extern	char *	tgetstr();
+	extern	int	tgetnum();
+	extern	int	tgetflag();
+# endif
+# ifdef HAVE_TPARM
+	extern	char *	tparm();
+# endif
 #endif
 
-extern	char	*getenv();
-extern	char	*tparm();
+/*
+ * We use macro wrappers to hide the difference between terminfo and termcap.
+ */
+#ifdef HAVE_TERMINFO
+# define Tgetstr(x, y) 	tigetstr(x.iname)
+# define Tgetnum(x) 	tigetnum(x.iname);
+# define Tgetflag(x) 	tigetflag(x.iname);
+#else
+# define Tgetstr(x, y) 	tgetstr(x.tname, &y)
+# define Tgetnum(x) 	tgetnum(x.tname)
+# define Tgetflag(x) 	tgetflag(x.tname)
+#endif
+
+/*
+ * And for the final inanity, X/OPEN curses uses
+ *	char *tparm (char *, int, int, int, int, int, int, int, int, int);
+ * but everyone else (including ncurses and solaris) uses
+ *	char *tparm (const char *, ...);
+ * and there's no way to cover those two arglists with one prototype, so
+ * we're left with these idiot wrapper functions.
+ */
+char *tparm1 (const char *str, int l1)
+{
+	return tparm(str, l1, 0, 0, 0, 0, 0, 0, 0, 0);
+}
+
+char *tparm2 (const char *str, int l1, int l2)
+{
+	return tparm(str, l1, l2, 0, 0, 0, 0, 0, 0, 0);
+}
+
+char *tparm4 (const char *str, int l1, int l2, int l3, int l4)
+{
+	return tparm(str, l1, l2, l3, l4, 0, 0, 0, 0, 0);
+}
+
 
 /*
  * The old code assumed termcap. termcap is almost always present, but on
@@ -111,7 +163,7 @@ typedef struct cap2info
 	void *		ptr;
 } cap2info;
 
-struct	term	TI;
+struct	my_term	TI;
 
 cap2info tcaps[] =
 {
@@ -615,7 +667,7 @@ cap2info tcaps[] =
 };
 
 
-struct term *	current_term = &TI;
+struct my_term *current_term = &TI;
 static	int	numcaps = sizeof(tcaps) / sizeof(cap2info);
 static	int	term_echo_flag = 1;
 static	int	li;
@@ -701,7 +753,7 @@ void	term_reset (void)
 	tcsetattr(tty_des, TCSADRAIN, &oldb);
 
 	if (current_term->TI_csr)
-		tputs_x(tparm(current_term->TI_csr, 0, main_screen->li - 1));
+		tputs_x(tparm2(current_term->TI_csr, 0, main_screen->li - 1));
 	term_gotoxy(0, main_screen->li - 1);
 #if use_alt_screen
 	if (current_term->TI_rmcup)
@@ -812,28 +864,12 @@ int 	term_init (void)
 		{
 			cval = Tgetstr(tcaps[i], tptr);
 			if (cval == (char *) -1)
-				*(char **)tcaps[i].ptr = NULL;
+				*(char * *)tcaps[i].ptr = NULL;
 			else
-				*(char **)tcaps[i].ptr = cval;
+				*(char * *)tcaps[i].ptr = cval;
 		}
 	}
 
-	BC = current_term->TI_cub1;
-	UP = current_term->TI_cuu1;
-	if (current_term->TI_pad)
-		my_PC = current_term->TI_pad[0];
-	else
-		my_PC = 0;
-
-	if (BC)
-		BClen = strlen(BC);
-	else
-		BClen = 0;
-	if (UP)
-		UPlen = strlen(UP);
-	else
-		UPlen = 0;
-
 	li = current_term->TI_lines;
 	co = current_term->TI_cols;
 	if (!co)
@@ -848,28 +884,28 @@ int 	term_init (void)
 
 	current_term->TI_normal[0] = 0;
 	if (current_term->TI_sgr0)
-		strcat(current_term->TI_normal, current_term->TI_sgr0);
+		strlcat(current_term->TI_normal, current_term->TI_sgr0, sizeof current_term->TI_normal);
 	if (current_term->TI_rmso)
 	{
 		if (current_term->TI_sgr0 && 
 		    strcmp(current_term->TI_rmso, current_term->TI_sgr0))
-			strcat(current_term->TI_normal, current_term->TI_rmso);
+			strlcat(current_term->TI_normal, current_term->TI_rmso, sizeof current_term->TI_sgr0);
 	}
 	if (current_term->TI_rmul)
 	{
 		if (current_term->TI_sgr0 && 
 		    strcmp(current_term->TI_rmul, current_term->TI_sgr0))
-			strcat (current_term->TI_normal, current_term->TI_rmul);
+			strlcat(current_term->TI_normal, current_term->TI_rmul, sizeof current_term->TI_normal);
 	}
 	/*
 	 * On some systems, alternate char set mode isn't exited with sgr0
 	 */
 	if (current_term->TI_rmacs)
-		strcat(current_term->TI_normal, current_term->TI_rmacs);
+		strlcat(current_term->TI_normal, current_term->TI_rmacs, sizeof current_term->TI_normal);
 
 
 	/*
-	 * Finally set up the current_term->TI_sgrstrs array.  
+	 * Finally set up the current_term->TI_sgrstrs array.
 	 * Clean it out first...
 	 */
 	for (i = 0; i < TERM_SGR_MAXVAL; i++)
@@ -1005,31 +1041,38 @@ int 	term_init (void)
 
 		*cbuf = 0;
 		if (current_term->TI_setaf) 
-		    strcat(cbuf, tparm(current_term->TI_setaf, i & 0x07, 0));
+		    strlcat(cbuf, tparm2(current_term->TI_setaf, i & 0x07, 0), sizeof cbuf);
 		else if (current_term->TI_setf)
-		    strcat(cbuf, tparm(current_term->TI_setf, i & 0x07, 0));
+		    strlcat(cbuf, tparm2(current_term->TI_setf, i & 0x07, 0), sizeof cbuf);
 		else
-		    sprintf(cbuf, "\033[%dm", (i & 0x07) + 30);
+		    snprintf(cbuf, sizeof cbuf, "\033[%dm", (i & 0x07) + 30);
 
-		current_term->TI_forecolors[i] = m_strdup(cbuf);
+		current_term->TI_forecolors[i] = malloc_strdup(cbuf);
 
 		*cbuf = 0;
 		if (current_term->TI_setab)
-		    strcat (cbuf, tparm(current_term->TI_setab, i & 0x07, 0));
+		    strlcat(cbuf, tparm2(current_term->TI_setab, i & 0x07, 0), sizeof cbuf);
 		else if (current_term->TI_setb)
-		    strcat (cbuf, tparm(current_term->TI_setb, i & 0x07, 0));
+		    strlcat(cbuf, tparm2(current_term->TI_setb, i & 0x07, 0), sizeof cbuf);
 		else
-			sprintf(cbuf, "\033[%dm", (i & 0x07) + 40);
+		    snprintf(cbuf, sizeof cbuf, "\033[%dm", (i & 0x07) + 40);
 
-		current_term->TI_backcolors[i] = m_strdup(cbuf);
+		current_term->TI_backcolors[i] = malloc_strdup(cbuf);
 	}
-            
+
 /* Set up the terminal discipline */
 	tty_des = 0;			/* Has to be. */
 	tcgetattr(tty_des, &oldb);
 
 	newb = oldb;
 	newb.c_lflag &= ~(ICANON | ECHO); /* set equ. of CBREAK, no ECHO */
+#ifdef IEXTEN
+	if (use_iexten == 1)
+		newb.c_lflag |= IEXTEN;	    /* Turn off DISCARD/LNEXT */
+	else if (use_iexten == 0)
+		newb.c_lflag &= ~(IEXTEN);  /* Turn off DISCARD/LNEXT */
+#endif
+
 	newb.c_cc[VMIN] = 1;	          /* read() satified after 1 char */
 	newb.c_cc[VTIME] = 0;	          /* No timer */
 
@@ -1041,6 +1084,10 @@ int 	term_init (void)
 #               endif
 #       endif
 
+	/* 
+	 * We can't just turn off ISIG since we want ^C to send SIGINT,
+	 * so we turn off all of the other signal-generating keys
+	 */
 	newb.c_cc[VQUIT] = _POSIX_VDISABLE;
 #	ifdef VDSUSP
 		newb.c_cc[VDSUSP] = _POSIX_VDISABLE;
@@ -1140,7 +1187,7 @@ void	term_beep (void)
 	}
 }
 
-int	orig_term_eight_bit (void)
+static int	orig_term_eight_bit (void)
 {
 	if (dumb_mode)
 		return 1;
@@ -1172,8 +1219,10 @@ void	set_term_eight_bit (int value)
 	tcsetattr(tty_des, TCSADRAIN, &newb);
 }
 
-void	set_meta_8bit (int value)
+void	set_meta_8bit (const void *stuff)
 {
+	int	value = *(const int *)stuff;
+
 	if (dumb_mode)
 		return;
 
@@ -1189,11 +1238,11 @@ void	set_meta_8bit (int value)
 void	term_gotoxy (int col, int row)
 {
 	if (current_term->TI_cup)
-		tputs_x(tparm(current_term->TI_cup, row, col));
+		tputs_x(tparm2(current_term->TI_cup, row, col));
 	else
 	{
-		tputs_x(tparm(current_term->TI_hpa, col));
-		tputs_x(tparm(current_term->TI_vpa, row));
+		tputs_x(tparm1(current_term->TI_hpa, col));
+		tputs_x(tparm1(current_term->TI_vpa, row));
 	}
 }
 
@@ -1220,7 +1269,7 @@ void	term_clrscr (void)
 	/* We can also clear by deleteing lines ... */
 	else if (current_term->TI_dl)
 	{
-		tputs_x(tparm(current_term->TI_dl, current_term->TI_lines));
+		tputs_x(tparm1(current_term->TI_dl, current_term->TI_lines));
 		return;
 	}
 	/* ... in this case one line at a time */
@@ -1233,7 +1282,7 @@ void	term_clrscr (void)
 	/* As a last resort we can insert lines ... */
 	else if (current_term->TI_il)
 	{
-		tputs_x (tparm(current_term->TI_il, current_term->TI_lines));
+		tputs_x (tparm1(current_term->TI_il, current_term->TI_lines));
 		term_gotoxy (0, 0);
 		return;
 	}
@@ -1254,9 +1303,9 @@ void	term_left (int num)
 	if (num == 1 && current_term->TI_cub1)
 		tputs_x(current_term->TI_cub1);
 	else if (current_term->TI_cub)
-		tputs_x (tparm(current_term->TI_cub, num));
+		tputs_x (tparm1(current_term->TI_cub, num));
 	else if (current_term->TI_mrcup)
-		tputs_x (tparm(current_term->TI_mrcup, -num, 0));
+		tputs_x (tparm2(current_term->TI_mrcup, -num, 0));
 	else if (current_term->TI_cub1)
 		while (num--)
 			tputs_x(current_term->TI_cub1);
@@ -1273,9 +1322,9 @@ void	term_right (int num)
 	if (num == 1 && current_term->TI_cuf1)
 		tputs_x(current_term->TI_cuf1);
 	else if (current_term->TI_cuf)
-		tputs_x (tparm(current_term->TI_cuf, num));
+		tputs_x (tparm1(current_term->TI_cuf, num));
 	else if (current_term->TI_mrcup)
-		tputs_x (tparm(current_term->TI_mrcup, num, 0));
+		tputs_x (tparm2(current_term->TI_mrcup, num, 0));
 	else if (current_term->TI_cuf1)
 		while (num--)
 			tputs_x(current_term->TI_cuf1);
@@ -1291,7 +1340,7 @@ void	term_delete (int num)
 		tputs_x(current_term->TI_smdc);
 
 	if (current_term->TI_dch)
-		tputs_x (tparm (current_term->TI_dch, num));
+		tputs_x (tparm1(current_term->TI_dch, num));
 	else if (current_term->TI_dch1)
 		while (num--)
 			tputs_x (current_term->TI_dch1);
@@ -1310,7 +1359,7 @@ void	term_insert (unsigned char c)
 	else if (current_term->TI_ich1)
 		tputs_x (current_term->TI_ich1);
 	else if (current_term->TI_ich)
-		tputs_x (tparm(current_term->TI_ich, 1));
+		tputs_x (tparm1(current_term->TI_ich, 1));
 
 	term_putchar (c);
 
@@ -1324,7 +1373,7 @@ void	term_insert (unsigned char c)
 void	term_repeat (unsigned char c, int rep)
 {
 	if (current_term->TI_rep)
-		tputs_x(tparm (current_term->TI_rep, (int)c, rep));
+		tputs_x(tparm2(current_term->TI_rep, (int)c, rep));
 	else
 		while (rep--)
 			putchar_x (c);
@@ -1363,8 +1412,8 @@ void	term_scroll (int top, int bot, int n)
 		 * region was the full screen.  That test *always* fails,
 		 * because we never scroll the bottom line of the screen.
 		 */
-		strcpy(start, tparm(current_term->TI_csr, top, bot));
-		strcpy(final, tparm(current_term->TI_csr, 0, current_term->TI_lines-1));
+		strlcpy(start, tparm2(current_term->TI_csr, top, bot), sizeof start);
+		strlcpy(final, tparm2(current_term->TI_csr, 0, current_term->TI_lines-1), sizeof final);
 
 		if (n > 0)
 		{
@@ -1373,10 +1422,10 @@ void	term_scroll (int top, int bot, int n)
 			if (current_term->TI_indn)
 			{
 				oneshot = 1;
-				strcpy(thing, tparm(current_term->TI_indn, rn, rn));
+				strlcpy(thing, tparm2(current_term->TI_indn, rn, rn), sizeof thing);
 			}
 			else
-				strcpy(thing, current_term->TI_ind);
+				strlcpy(thing, current_term->TI_ind, sizeof thing);
 		}
 		else
 		{
@@ -1385,17 +1434,17 @@ void	term_scroll (int top, int bot, int n)
 			if (current_term->TI_rin)
 			{
 				oneshot = 1;
-				strcpy (thing, tparm(current_term->TI_rin, rn, rn));
+				strlcpy(thing, tparm2(current_term->TI_rin, rn, rn), sizeof thing);
 			}
 			else
-				strcpy (thing, current_term->TI_ri);
+				strlcpy(thing, current_term->TI_ri, sizeof thing);
 		}
 	}
 
 	else if (current_term->TI_wind && (current_term->TI_ri || current_term->TI_rin) && (current_term->TI_ind || current_term->TI_indn))
 	{
-		strcpy(start, tparm(current_term->TI_wind, top, bot, 0, current_term->TI_cols-1));
-		strcpy(final, tparm(current_term->TI_wind, 0, current_term->TI_lines-1, 0, current_term->TI_cols-1));
+		strlcpy(start, tparm4(current_term->TI_wind, top, bot, 0, current_term->TI_cols-1), sizeof start);
+		strlcpy(final, tparm4(current_term->TI_wind, 0, current_term->TI_lines-1, 0, current_term->TI_cols-1), sizeof final);
 
 		if (n > 0)
 		{
@@ -1404,10 +1453,10 @@ void	term_scroll (int top, int bot, int n)
 			if (current_term->TI_indn)
 			{
 				oneshot = 1;
-				strcpy (thing, tparm(current_term->TI_indn, rn, rn));
+				strlcpy(thing, tparm2(current_term->TI_indn, rn, rn), sizeof thing);
 			}
 			else
-				strcpy (thing, current_term->TI_ind);
+				strlcpy(thing, current_term->TI_ind, sizeof thing);
 		}
 		else
 		{
@@ -1416,10 +1465,10 @@ void	term_scroll (int top, int bot, int n)
 			if (current_term->TI_rin)
 			{
 				oneshot = 1;
-				strcpy (thing,tparm(current_term->TI_rin, rn, rn));
+				strlcpy(thing, tparm2(current_term->TI_rin, rn, rn), sizeof thing);
 			}
 			else
-				strcpy (thing, current_term->TI_ri);
+				strlcpy(thing, current_term->TI_ri, sizeof thing);
 		}
 	}
 
@@ -1433,18 +1482,18 @@ void	term_scroll (int top, int bot, int n)
 			if (current_term->TI_dl)
 			{
 				oneshot = 1;
-				strcpy (thing, tparm(current_term->TI_dl, rn, rn));
+				strlcpy(thing, tparm2(current_term->TI_dl, rn, rn), sizeof thing);
 			}
 			else
-				strcpy (thing, current_term->TI_dl1);
+				strlcpy(thing, current_term->TI_dl1, sizeof thing);
 
 			if (current_term->TI_il)
 			{
 				oneshot = 1;
-				strcpy(final, tparm(current_term->TI_il, rn, rn));
+				strlcpy(final, tparm2(current_term->TI_il, rn, rn), sizeof final);
 			}
 			else
-				strcpy(final, current_term->TI_il1);
+				strlcpy(final, current_term->TI_il1, sizeof final);
 		}
 		else
 		{
@@ -1453,18 +1502,18 @@ void	term_scroll (int top, int bot, int n)
 			if (current_term->TI_il)
 			{
 				oneshot = 1;
-				strcpy (thing, tparm(current_term->TI_il, rn, rn));
+				strlcpy(thing, tparm2(current_term->TI_il, rn, rn), sizeof thing);
 			}
 			else
-				strcpy (thing, current_term->TI_il1);
+				strlcpy(thing, current_term->TI_il1, sizeof thing);
 
 			if (current_term->TI_dl)
 			{
 				oneshot = 1;
-				strcpy(final, tparm(current_term->TI_dl, rn, rn));
+				strlcpy(final, tparm2(current_term->TI_dl, rn, rn), sizeof thing);
 			}
 			else
-				strcpy(final, current_term->TI_dl1);
+				strlcpy(final, current_term->TI_dl1, sizeof thing);
 		}
 	}
 
@@ -1526,9 +1575,9 @@ void	term_scroll (int top, int bot, int n)
  * 14   - bright cyan (foreground only)
  * 15   - bright white (foreground only)
  */
-char *	term_getsgr (int opt, int fore, int back)
+const char *	term_getsgr (int opt, int fore, int back)
 {
-	char *ret = empty_string;
+	const char *ret = empty_string;
 
 	switch (opt)
 	{
@@ -1552,7 +1601,7 @@ char *	term_getsgr (int opt, int fore, int back)
 			break;
 		case TERM_SGR_GCHAR:
 			if (current_term->TI_dispc)
-				ret = tparm(current_term->TI_dispc, fore);
+				ret = tparm1(current_term->TI_dispc, fore);
 			break;
 		default:
 			panic ("Unknown option '%d' to term_getsgr", opt);
@@ -1561,7 +1610,7 @@ char *	term_getsgr (int opt, int fore, int back)
 	return (ret);
 }
 
-char *	control_mangle (unsigned char *text)
+static char *	control_mangle (unsigned char *text)
 {
 static 	u_char	retval[256];
 	int 	pos = 0;
@@ -1590,7 +1639,7 @@ static 	u_char	retval[256];
 	return retval;
 }
 
-char *	get_term_capability (char *name, int querytype, int mangle)
+const char *	get_term_capability (const char *name, int querytype, int mangle)
 {
 static	char		retval[128];
 	const char *	compare = empty_string;
@@ -1615,14 +1664,14 @@ static	char		retval[128];
 			case CAP_TYPE_INT:
 				if (!(int *)t->ptr)
 					return NULL;
-				strcpy(retval, ltoa(* (int *)(t->ptr)));
+				strlcpy(retval, ltoa(* (int *)(t->ptr)), sizeof retval);
 				return retval;
 			case CAP_TYPE_STR:
 				if (!(char **)t->ptr || !*(char **)t->ptr)
 					return NULL;
-				strcpy(retval, mangle ? 
+				strlcpy(retval, mangle ? 
 					control_mangle(*(char **)t->ptr) :
-						      (*(char **)t->ptr));
+						      (*(char **)t->ptr), sizeof retval);
 				return retval;
 		    }
 		}
